/*
 * gtkimcontexttim.c - Table-based Input Method for GTK+ 2 class.
 *
    Copyright (C) 2002-2003  Yao Zhang

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * yaoz@users.sourceforge.net
 */
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gdk/gdkkeysyms.h>

#include "gtkimcontexttim.h"
#include "timtable.h"


static GHashTable *TABLES;

#undef DEBUG

/**
 * Locate one UTF-8 character sequence at the current location in string.
 *
 * The search starts from current location "cp".  First, "cp" has to be in
 * string, that it, cp >= s and cp <= s+strlen(s).  Note that "cp" could be
 * at the trailing '\0'.
 *
 * Then search backward until a valid UTF-8 character leading byte
 * is obtained.  But the search doesn't go beyond 6 bytes (maximum length
 * of a UTF-8 sequence).  If we DO have a leading byte, we also check if
 * the following bytes are valid or not.
 *
 * It returns the length of the valid UTF-8 sequence for the current
 * character and "cp" will be repositioned to point to the leading byte.
 * If "cp" is not within a valid UTF-8 character sequence, the function
 * returns 0 and "cp" is not changed.
 *
 * To use this function, "&cp" is passed as the second parameter "acp" (address
 * of char pointer). If "acp" is NULL, then "cp" is set to "s".
 */
static int strutf8seq(const guchar *s, const guchar **acp)
{
    int length;
    int count;
    const guchar *cp, *p;

    if (s == NULL) {
        return 0;
    }

    if (acp == NULL) {
        cp = s;
    } else {
        cp = *acp;
    }

    if (cp < s) {
        return 0;
    }

    length = strlen((char *)s);
    if ((cp - s) > length) {
        return 0;
    }

    count = 0;
    p = cp;
    while (p >= s) {
        if (((*p)>>6) == 0x02) {
            /* Following byte, back off. */
            if ((cp - p) < 5) {
                p--;
                continue;
            }
        } else if (((*p)>>7) == 0x0) {
            count = 1;
        } else if (((*p)>>5) == 0x06) {
            count = 2;
        } else if (((*p)>>4) == 0x0e) {
            count = 3;
        } else if (((*p)>>3) == 0x1e) {
            count = 4;
        } else if (((*p)>>2) == 0x3e) {
            count = 5;
        } else if (((*p)>>1) == 0x7e) {
            count = 6;
        }
        break;
    }
    if (count > 1) {
        int i;

        for (i = 1; i < count; i++) {
            if (((*(p+i))>>6) != 0x02) {
                return 0;
            }
        }
        if (acp != NULL) {
            *acp = p;
        }
    }

    return count;
}

static TIMTable *get_table(GtkIMContextTIM *this)
{
    return (TIMTable *)g_hash_table_lookup(TABLES, this->id);
}

/**
 * Move candidate window based on cursor position and size.
 * Note that cursor position here is in root window coordinate.
 */
static void candidate_window_move(GtkIMContextTIM *this,
                                  int cursor_x, int cursor_y,
                                  int cursor_xs, int cursor_ys)
{
    int x, y;

    gtk_window_get_size(GTK_WINDOW(this->candidate_window),
                        &this->xs, &this->ys);

    x = cursor_x + cursor_xs/2 - this->xs/4;
    /* Reduce the horizontal position change. */
    if (((x-this->x) > (-this->xs/4))
     && ((x-this->x) < (+this->xs/4))) {
        x = this->x;
    }
    if (x < 8) {
        x = 8;
    } else if (x+this->xs > gdk_screen_width()-8) {
        x = gdk_screen_width()-8 - this->xs;
    }

    y = cursor_y + cursor_ys + 8;
    if (y < 8) {
        y = 8;
    } else if (y+this->ys > gdk_screen_height()-8) {
        y = gdk_screen_height()-8 - this->ys;
    }
    if ((y < cursor_y+cursor_ys) && (y+this->ys > cursor_y)) {
        y = cursor_y-8 - this->ys;
    }

    this->x = x;
    this->y = y;

    if (this->showing) {
        gtk_window_move(GTK_WINDOW(this->candidate_window), x, y);
    }
}

enum tim_actions {
    TIM_KEYIN,
    TIM_SWITCH,
    TIM_NEXT,
    TIM_PREVIOUS,
    TIM_NEXT_PAGE,
    TIM_PREVIOUS_PAGE,
    TIM_SELECT,
    TIM_SELECT1,
    TIM_SELECT2,
    TIM_SELECT3,
    TIM_SELECT4,
    TIM_SELECT5,
    TIM_SELECT6,
    TIM_SELECT7,
    TIM_SELECT8,
    TIM_SELECT9,
    TIM_SELECT0,
    TIM_COMMIT,
    TIM_BACKSPACE,
    TIM_END
};

static int tim_action(guint keyval, guint state)
{
    int n;

    n = TIM_KEYIN;

    if ((state&(GDK_SHIFT_MASK)) != 0) {
        switch (keyval) {
        case GDK_space:
            n = TIM_SWITCH;
            break;
        default:
            break;
        }
    }

    if (n != TIM_KEYIN) {
        return n;
    }

    switch (keyval) {
    case GDK_Escape:
        n = TIM_SWITCH;
        break;
    case GDK_comma:
        n = TIM_PREVIOUS;
        break;
    case GDK_period:
        n = TIM_NEXT;
        break;
    case GDK_less:
        n = TIM_PREVIOUS_PAGE;
        break;
    case GDK_greater:
        n = TIM_NEXT_PAGE;
        break;
    case GDK_space:
        n = TIM_SELECT;
        break;
    case GDK_1:
        n = TIM_SELECT1;
        break;
    case GDK_2:
        n = TIM_SELECT2;
        break;
    case GDK_3:
        n = TIM_SELECT3;
        break;
    case GDK_4:
        n = TIM_SELECT4;
        break;
    case GDK_5:
        n = TIM_SELECT5;
        break;
    case GDK_6:
        n = TIM_SELECT6;
        break;
    case GDK_7:
        n = TIM_SELECT7;
        break;
    case GDK_8:
        n = TIM_SELECT8;
        break;
    case GDK_9:
        n = TIM_SELECT9;
        break;
    case GDK_0:
        n = TIM_SELECT0;
        break;
    case GDK_Return:
        n = TIM_COMMIT;
        break;
    case GDK_BackSpace:
        n = TIM_BACKSPACE;
        break;
    default:
        break;
    }

    return n;
}

static int handle_keyin(GtkIMContextTIM *tim, GdkEventKey *event)
{
    int c;

    c = '\0';
    if ((event->keyval >= GDK_space)
     && (event->keyval <= GDK_asciitilde)) {
        c = event->keyval;
    } else if (event->keyval == GDK_Return) {
        c = '\n';
    }

    if (c != '\0') {
        g_string_append_c(tim->raw, c);
        g_string_set_size(tim->candidates, 0);
        tim->which = NULL;
        g_signal_emit_by_name(tim, "preedit_changed");

        return TRUE;
    }

    return FALSE;
}

static int handle_switch(GtkIMContextTIM *tim, GdkEventKey *event)
{
    register int i;
    register GString *s;

    s = g_string_new(tim->converted->str);
    i = strutf8seq((guchar *)tim->candidates->str, &tim->which);
    if (i > 0) {
        g_string_append_len(s, (gchar *)tim->which, i);
    } else {
        g_string_append(s, tim->raw->str);
    }
    g_signal_emit_by_name(tim, "commit", s->str);
    g_string_free(s, TRUE);
    g_string_set_size(tim->converted, 0);
    g_string_set_size(tim->raw, 0);
    g_string_set_size(tim->candidates, 0);
    tim->which = NULL;
    g_signal_emit_by_name(tim, "preedit_changed");
    tim->ascii_mode = TRUE;

    return TRUE;
}

static void next_n(GtkIMContextTIM *tim, int n)
{
    if ((tim->candidates->len > 0)
     && (strchr(tim->candidates->str, US) != NULL)) {
        while (n--) {
            register int i;

            i = strutf8seq((guchar *)tim->candidates->str, &tim->which);
            if (i > 0) {
                const guchar *p;

                p = (guchar *)strchr((char *)tim->which, US);
                tim->which = (guchar *)tim->candidates->str;
                if (p != NULL) {
                    p++;
                    i = strutf8seq((guchar *)tim->candidates->str, &p);
                    if (i > 0) {
                        if (*p != '\0') {
                            tim->which = p;
                        }
                    }
                }
            }
        }
    }
}

static void previous_n(GtkIMContextTIM *tim, int n)
{
    if ((tim->candidates->len > 0)
     && (strchr(tim->candidates->str, US) != NULL)) {
        while (n--) {
            register int i;

            i = strutf8seq((guchar *)tim->candidates->str, &tim->which);
            if (i > 0) {
                const guchar *p;

                p = (guchar *)strrchr(tim->candidates->str, US);
                if (p != NULL) {
                    const guchar *q;

                    q = tim->which - 2;
                    if (q < (guchar *)tim->candidates->str) {
                        p++;
                        i = strutf8seq((guchar *)tim->candidates->str, &p);
                        if (i > 0) {
                            if (*p != '\0') {
                                tim->which = p;
                            }
                        }
                    } else {
                        tim->which = (guchar *)tim->candidates->str;
                        while (q >= (guchar *)tim->candidates->str) {
                            if (*q == US) {
                                q++;
                                i = strutf8seq((guchar *)tim->candidates->str, &q);
                                if (i > 0) {
                                    if (*q != '\0') {
                                        tim->which = q;
                                    }
                                }
                                break;
                            }
                            q--;
                        }
                    }
                }
            }
        }
    }
}

static int handle_next(GtkIMContextTIM *tim, GdkEventKey *event)
{
    if ((tim->candidates->len > 0)
     && (strchr(tim->candidates->str, US) != NULL)) {
        next_n(tim, 1);
    } else {
        handle_keyin(tim, event);
    }

    return TRUE;
}

static int handle_previous(GtkIMContextTIM *tim, GdkEventKey *event)
{
    if ((tim->candidates->len > 0)
     && (strchr(tim->candidates->str, US) != NULL)) {
        previous_n(tim, 1);
    } else {
        handle_keyin(tim, event);
    }

    return TRUE;
}

static int handle_next_page(GtkIMContextTIM *tim, GdkEventKey *event)
{
    if ((tim->candidates->len > 0)
     && (strchr(tim->candidates->str, US) != NULL)) {
        next_n(tim, 10);
    } else {
        handle_keyin(tim, event);
    }

    return TRUE;
}

static int handle_previous_page(GtkIMContextTIM *tim, GdkEventKey *event)
{
    if ((tim->candidates->len > 0)
     && (strchr(tim->candidates->str, US) != NULL)) {
        previous_n(tim, 10);
    } else {
        handle_keyin(tim, event);
    }

    return TRUE;
}

static int handle_select(GtkIMContextTIM *tim, GdkEventKey *event)
{
    register int i;

    i = strutf8seq((guchar *)tim->candidates->str, &tim->which);
    if (i > 0) {
        guchar *p;

        p = (guchar *)strchr((char *)tim->which, US);
        if (p != NULL) {
            i = p - tim->which;
        } else {
            i = strlen((char *)tim->which);
        }
        g_string_append_len(tim->converted, (gchar *)tim->which, i);
    } else if (tim->raw->len > 0) {
        g_string_append(tim->converted, tim->raw->str);
    } else {
        if ((event->keyval >= GDK_space)
         && (event->keyval <= GDK_asciitilde)) {
            g_string_append_c(tim->converted, event->keyval);
        } else if (event->keyval == GDK_Return) {
            g_string_append_c(tim->converted, '\n');
        }
    }
    g_string_set_size(tim->raw, 0);
    g_string_set_size(tim->candidates, 0);
    tim->which = NULL;
    g_signal_emit_by_name(tim, "preedit_changed");

    return TRUE;
}

static int handle_select_n(GtkIMContextTIM *tim, GdkEventKey *event, int n)
{
    guchar *p;

    p = NULL;

    if ((tim->candidates->len > 0)
     && (strchr(tim->candidates->str, US) != NULL)) {
        register int i;
        GString *s;

        s = g_string_new((gchar *)tim->which);
        if ((gchar *)tim->which != tim->candidates->str) {
            g_string_append_c(s, US);
            g_string_append_len(s, tim->candidates->str,
                    (gchar *)tim->which-1 - tim->candidates->str);
        }

        p = (guchar *)s->str;
        for (i = 0; i < n; i++) {
            p = (guchar *)strchr((char *)p, US);
            if (p != NULL) {
                p++;
            } else {
                break;
            } 
        }

        if (p != NULL) {
            guchar *q;

            q = (guchar *)strchr((char *)p, US);
            if (q != NULL) {
                i = q - p;
            } else {
                i = strlen((char *)p);
            }
            g_string_append_len(tim->converted, (gchar *)p, i);
            g_string_set_size(tim->raw, 0);
            g_string_set_size(tim->candidates, 0);
            tim->which = NULL;
            g_signal_emit_by_name(tim, "preedit_changed");
        }

        g_string_free(s, TRUE);
    }

    if (p == NULL) {
        handle_keyin(tim, event);
    }

    return TRUE;
}

static int handle_select1(GtkIMContextTIM *tim, GdkEventKey *event)
{
    return handle_select_n(tim, event, 1);
}

static int handle_select2(GtkIMContextTIM *tim, GdkEventKey *event)
{
    return handle_select_n(tim, event, 2);
}

static int handle_select3(GtkIMContextTIM *tim, GdkEventKey *event)
{
    return handle_select_n(tim, event, 3);
}

static int handle_select4(GtkIMContextTIM *tim, GdkEventKey *event)
{
    return handle_select_n(tim, event, 4);
}

static int handle_select5(GtkIMContextTIM *tim, GdkEventKey *event)
{
    return handle_select_n(tim, event, 5);
}

static int handle_select6(GtkIMContextTIM *tim, GdkEventKey *event)
{
    return handle_select_n(tim, event, 6);
}

static int handle_select7(GtkIMContextTIM *tim, GdkEventKey *event)
{
    return handle_select_n(tim, event, 7);
}

static int handle_select8(GtkIMContextTIM *tim, GdkEventKey *event)
{
    return handle_select_n(tim, event, 8);
}

static int handle_select9(GtkIMContextTIM *tim, GdkEventKey *event)
{
    return handle_select_n(tim, event, 9);
}

static int handle_select0(GtkIMContextTIM *tim, GdkEventKey *event)
{
    return handle_select_n(tim, event, 10);
}

static int handle_commit(GtkIMContextTIM *tim, GdkEventKey *event)
{
    register int i;
    GString *s;

    s = g_string_new(tim->converted->str);
    i = strutf8seq((guchar *)tim->candidates->str, &tim->which);
    if (i > 0) {
        guchar *p;

        p = (guchar *)strchr((char *)tim->which, US);
        if (p != NULL) {
            i = p - tim->which;
        } else {
            i = strlen((char *)tim->which);
        }
        g_string_append_len(s, (gchar *)tim->which, i);
    } else {
        g_string_append(s, tim->raw->str);
    }
    if (s->len <= 0) {
        if ((event->keyval >= GDK_space)
         && (event->keyval <= GDK_asciitilde)) {
            g_string_append_c(s, event->keyval);
        } else if (event->keyval == GDK_Return) {
            g_string_append_c(s, '\n');
        }
    }
    g_signal_emit_by_name(tim, "commit", s->str);
    g_string_free(s, TRUE);
    g_string_set_size(tim->converted, 0);
    g_string_set_size(tim->raw, 0);
    g_string_set_size(tim->candidates, 0);
    tim->which = NULL;
    g_signal_emit_by_name(tim, "preedit_changed");

    return TRUE;
}

static int handle_backspace(GtkIMContextTIM *tim, GdkEventKey *event)
{
    register int i;
    TIMTable *table;
    GString *s;

    table = get_table(tim);
    if (table == NULL) {
	return FALSE;
    }

    if (tim->raw->len > 0) {
        g_string_truncate(tim->raw, tim->raw->len - 1);
        s = TIMTABLE_GET_CLASS(table)->lookup(table, tim->raw);
        if (s != NULL) {
            g_string_assign(tim->candidates, s->str);
            tim->which = (guchar *)tim->candidates->str;
        } else {
            g_string_set_size(tim->candidates, 0);
            tim->which = NULL;
        }
    } else if (tim->converted->len > 0) {
        const guchar *p;

        p = (guchar *)(tim->converted->str + tim->converted->len - 1);
        i = strutf8seq((guchar *)tim->converted->str, &p);
        if (i > 0) {
            g_string_truncate(tim->converted,
                              tim->converted->len - i);
        }
    } else {
	return FALSE;
    }
    g_signal_emit_by_name(tim, "preedit_changed");

    return TRUE;
}

static struct {
    enum tim_actions action;
    int (*handler)(GtkIMContextTIM *tim, GdkEventKey *event);
} HANDLERS[] = {
    {TIM_KEYIN, handle_keyin},
    {TIM_SWITCH, handle_switch},
    {TIM_NEXT, handle_next},
    {TIM_PREVIOUS, handle_previous},
    {TIM_NEXT_PAGE, handle_next_page},
    {TIM_PREVIOUS_PAGE, handle_previous_page},
    {TIM_SELECT, handle_select},
    {TIM_SELECT1, handle_select1},
    {TIM_SELECT2, handle_select2},
    {TIM_SELECT3, handle_select3},
    {TIM_SELECT4, handle_select4},
    {TIM_SELECT5, handle_select5},
    {TIM_SELECT6, handle_select6},
    {TIM_SELECT7, handle_select7},
    {TIM_SELECT8, handle_select8},
    {TIM_SELECT9, handle_select9},
    {TIM_SELECT0, handle_select0},
    {TIM_COMMIT, handle_commit},
    {TIM_BACKSPACE, handle_backspace},
    {TIM_END, NULL}
};

static void process_candidates(GString *s)
{
    register int i;
    register gchar *p;
    gchar buffer[32];

    strcpy(buffer, "<sup><b><i> </i></b></sup>");
    g_string_insert(s, 0, buffer);
    for (i = 1; i <= 10; i++) {
        p = strchr(s->str, US);
        if (p != NULL) {
            *p = ' ';
            buffer[11] = (i%10) + '0';
            g_string_insert(s, p+1 - s->str, buffer);
            
        } else {
            break;
        }
    }

    while ((p = strchr(s->str, US)) != NULL) {
        *p = ' ';
    }

    strcpy(buffer, "<big>");
    g_string_prepend(s, buffer);
    strcpy(buffer, "</big>");
    g_string_append(s, buffer);
}


static gboolean candidate_window_expose_callback(GtkWidget *widget,
                                                 GdkEventExpose *event,
                                                 gpointer data)
{
    GtkRequisition requisition;

    gtk_widget_size_request(widget, &requisition);
    gtk_paint_flat_box(widget->style, widget->window,
                       GTK_STATE_NORMAL, GTK_SHADOW_NONE,
                       NULL, widget, "tooltip",
                       0, 0, requisition.width, requisition.height);

  return FALSE;
}

static void gtkimcontexttim_init(GTypeInstance *this, gpointer class)
{
    if (TABLES == NULL) {
        TABLES = g_hash_table_new_full((GHashFunc)g_str_hash,
                                       (GEqualFunc)g_str_equal,
                                       (GDestroyNotify)g_free,
                                       NULL);
    }

    GTKIMCONTEXTTIM(this)->debug = 0;
    if (getenv("WJ_DEBUG") != NULL) {
        GTKIMCONTEXTTIM(this)->debug = 1;
    }
    if (GTKIMCONTEXTTIM(this)->debug) {
        printf("gtkimcontexttim_init\n");
    }

    GTKIMCONTEXTTIM(this)->carryover_state = 0;
    GTKIMCONTEXTTIM(this)->client_window = NULL;
    GTKIMCONTEXTTIM(this)->ascii_mode = FALSE;
    GTKIMCONTEXTTIM(this)->converted = g_string_new("");
    GTKIMCONTEXTTIM(this)->raw = g_string_new("");
    GTKIMCONTEXTTIM(this)->candidates = g_string_new("");
    GTKIMCONTEXTTIM(this)->which = NULL;
    GTKIMCONTEXTTIM(this)->candidate_window = gtk_window_new(GTK_WINDOW_POPUP);
    gtk_widget_set_app_paintable(GTKIMCONTEXTTIM(this)->candidate_window, TRUE);
    gtk_container_set_border_width(GTK_CONTAINER(GTKIMCONTEXTTIM(this)->candidate_window), 1);
    g_signal_connect(GTKIMCONTEXTTIM(this)->candidate_window,
                    "expose_event",
                    (GCallback)candidate_window_expose_callback,
                    this);
    GTKIMCONTEXTTIM(this)->candidate_label = gtk_label_new("");
    gtk_misc_set_alignment(GTK_MISC(GTKIMCONTEXTTIM(this)->candidate_label),
                           0, 0.5);
    gtk_container_add(GTK_CONTAINER(GTKIMCONTEXTTIM(this)->candidate_window),
                      GTKIMCONTEXTTIM(this)->candidate_label);
/*
    g_signal_connect(GTKIMCONTEXTTIM(this)->candidate_window,
                    "destroy",
                    (GCallback)gtk_widget_destroyed,
                    &GTKIMCONTEXTTIM(this)->candidate_window);
*/
    GTKIMCONTEXTTIM(this)->showing = 0;
    GTKIMCONTEXTTIM(this)->x = 0;
    GTKIMCONTEXTTIM(this)->y = 0;
    GTKIMCONTEXTTIM(this)->xs = 384;
    GTKIMCONTEXTTIM(this)->ys = -1;
    gtk_widget_set_size_request(GTKIMCONTEXTTIM(this)->candidate_window,
                                GTKIMCONTEXTTIM(this)->xs,
                                GTKIMCONTEXTTIM(this)->ys);
    gtk_window_move(GTK_WINDOW(GTKIMCONTEXTTIM(this)->candidate_window),
                    64, 64);
    GTKIMCONTEXTTIM(this)->auto_commit = -1;
    GTKIMCONTEXTTIM(this)->id = NULL;
}

static void gtkimcontexttim_finalize(GObject *this)
{
    GtkIMContext *super_class;

    if (GTKIMCONTEXTTIM(this)->debug) {
        printf("gtkimcontexttim_finalize\n");
    }

    if (TABLES) {
	TIMTable *table;

	table = get_table(GTKIMCONTEXTTIM(this));
	if (table != NULL) {
            if (G_OBJECT(table)->ref_count == 1) {
	        g_hash_table_remove(TABLES, GTKIMCONTEXTTIM(this)->id);
	    }
	    g_object_unref(table);
	}
	if (g_hash_table_size(TABLES) <= 0) {
            if (GTKIMCONTEXTTIM(this)->debug) {
                printf("Destroying TABLES.\n");
            }
            g_hash_table_destroy(TABLES);
	    TABLES = NULL;
	}
    }

    /* GtkIMContextTIM specific finalization. */
    if (GTKIMCONTEXTTIM(this)->client_window != NULL) {
        g_object_unref(GTKIMCONTEXTTIM(this)->client_window);
    }
    g_string_free(GTKIMCONTEXTTIM(this)->converted, TRUE);
    g_string_free(GTKIMCONTEXTTIM(this)->raw, TRUE);
    g_string_free(GTKIMCONTEXTTIM(this)->candidates, TRUE);
/*
    g_signal_emit_by_name(GTKIMCONTEXTTIM(this)->candidate_window,
                          "destroy");
*/
    gtk_widget_destroy(GTKIMCONTEXTTIM(this)->candidate_window);
    if (GTKIMCONTEXTTIM(this)->id != NULL) {
        g_free(GTKIMCONTEXTTIM(this)->id);
    }

    /* Super class finalizatoin. */
    super_class = g_type_class_peek(g_type_parent(GTKIMCONTEXTTIM_TYPE));
    G_OBJECT_CLASS(super_class)->finalize(this);
}

static void gtkimcontexttim_set(GtkIMContextTIM *this, const gchar *id)
{
    if (GTKIMCONTEXTTIM(this)->debug) {
        printf("gtkimcontexttim_set\n");
    }

    if (this->id != NULL) {
        g_free(this->id);
        this->id = NULL;
    }
    if (id != NULL) {
        TIMTable *table;

        this->id = g_strdup(id);

        table = get_table(this);
        if (table == NULL) {
            table = g_object_new(TIMTABLE_TYPE, NULL);
            TIMTABLE_GET_CLASS(table)->load(table, "my.tim");
            TIMTABLE_GET_CLASS(table)->load(table, this->id);
            g_hash_table_insert(TABLES, g_strdup(this->id), table);
        } else {
	    g_object_ref(table);
	}
    }
}

static void gtkimcontexttim_set_client_window(GtkIMContext *this,
                                              GdkWindow *client_window)
{
    GtkIMContextTIM *this_tim = GTKIMCONTEXTTIM(this);

    if (GTKIMCONTEXTTIM(this)->debug) {
        printf("gtkimcontexttim_set_client_window\n");
    }

    if (this_tim->client_window != NULL) {
        g_object_unref(this_tim->client_window);
        this_tim->client_window = NULL;
    }
    if (client_window != NULL) {
        this_tim->client_window = g_object_ref(client_window);
    }
}

static void carryover_state(GtkIMContextTIM *this, GdkEventKey *event)
{
    guint state;

    state = event->state;
    event->state |= this->carryover_state;

    switch (event->keyval) {
    case GDK_Shift_L:
    case GDK_Shift_R:
        this->carryover_state |= state | GDK_SHIFT_MASK;
        break;
    case GDK_Control_L:
    case GDK_Control_R:
        this->carryover_state |= state | GDK_CONTROL_MASK;
        break;
    case GDK_Caps_Lock:
    case GDK_Shift_Lock:
        this->carryover_state |= state | GDK_LOCK_MASK;
        break;
    case GDK_Meta_L:
    case GDK_Meta_R:
        break;
    case GDK_Alt_L:
    case GDK_Alt_R:
        this->carryover_state |= state | GDK_MOD1_MASK;
        break;
    case GDK_Super_L: /* Windows */
    case GDK_Super_R:
        this->carryover_state |= state | GDK_MOD4_MASK;
        break;
    case GDK_Hyper_L:
    case GDK_Hyper_R:
        break;
    default:
        this->carryover_state = 0;
        break;
    }
}

static gboolean gtkimcontexttim_filter_keypress(GtkIMContext *this,
                                                GdkEventKey *event)
{
    register int handled;
    register GString *s;
    GtkIMContextTIM *this_tim = GTKIMCONTEXTTIM(this);

    if (GTKIMCONTEXTTIM(this)->debug) {
        printf("gtkimcontexttim_filter_keypress: state - 0x%X, keyval - 0x%X\n",
               event->state, event->keyval);
    }

    if (event->type != GDK_KEY_PRESS) {
        return FALSE;
    }

    /**
     * Allow special key state can be carried over to the next keypress
     * so that, for example, Ctrl+1 can be typed as press Ctrl key and
     * release it, then press 1 key.
     */
    carryover_state(this_tim, event);

    if (this_tim->ascii_mode) {
        if (((event->state&(GDK_SHIFT_MASK)) != 0)
         && (event->keyval == GDK_space)) {
            this_tim->ascii_mode = FALSE;
        } else {
            guint32 u;
            guchar utf8[16];

            u = gdk_keyval_to_unicode(event->keyval);
            if (u == 0) {
                return FALSE;
            }
            utf8[g_unichar_to_utf8(u, (gchar *)utf8)] = '\0';
            g_signal_emit_by_name(this_tim, "commit", utf8);
        }
        return TRUE;
    }

    handled = FALSE;
    /**
     * First thing's first: try to look it up
     * if NOT with special control keys like
     * Ctrl, Alt, etc.  Those special control
     * keys make the key combination processed
     * as command.
     */
    if (!(event->state&~(GDK_SHIFT_MASK|GDK_LOCK_MASK))
     && (event->keyval >= GDK_space)
     && (event->keyval <= GDK_asciitilde)) {
	TIMTable *table;

	table = get_table(this_tim);
	if (table != NULL) {
            GString *k;

            k = g_string_new(this_tim->raw->str);
            g_string_append_c(k, event->keyval);
            s = TIMTABLE_GET_CLASS(table)->lookup(table, k);
            if (s != NULL) {
                g_string_append_c(this_tim->raw, event->keyval);
                g_string_assign(this_tim->candidates, s->str);
                this_tim->which = (guchar *)this_tim->candidates->str;
                g_signal_emit_by_name(this_tim, "preedit_changed");
                handled = TRUE;
    	        /* AutoCommit */
    	        if (this_tim->auto_commit < 0) {
                    GString *key, *value;

                    key = g_string_new("AutoCommit");
                    value = TIMTABLE_GET_CLASS(table)->lookup(table, key);
                    GTKIMCONTEXTTIM(this)->auto_commit = 0;
                    if (value != NULL) {
                        GTKIMCONTEXTTIM(this)->auto_commit = atoi(value->str);
                    }
                    g_string_free(key, TRUE);
                }
        	if (this_tim->auto_commit && strchr(s->str, US) == NULL) {
        	    g_signal_emit_by_name(this_tim, "commit", s->str);
		    g_string_set_size(this_tim->converted, 0);
		    g_string_set_size(this_tim->raw, 0);
		    g_string_set_size(this_tim->candidates, 0);
		    this_tim->which = NULL;
		    g_signal_emit_by_name(this_tim, "preedit_changed");
	        }
            }
            g_string_free(k, TRUE);
	}
    }

    /* No lookup result.  Handle it as command. */
    if (!handled) {
        handled = HANDLERS[tim_action(event->keyval, event->state)].
			handler(this_tim, event);
    }

    this_tim->showing = 0;
    if (strchr(this_tim->candidates->str, US) != NULL) {
        this_tim->showing = 1;
    }
    if (this_tim->showing) {
        s = g_string_new((gchar *)this_tim->which);
        if ((gchar *)this_tim->which != this_tim->candidates->str) {
            g_string_append_c(s, US);
            g_string_append_len(s, this_tim->candidates->str,
                    (gchar *)this_tim->which-1 - this_tim->candidates->str);
        }
        process_candidates(s);
        gtk_label_set_markup(GTK_LABEL(this_tim->candidate_label), s->str);
        g_string_free(s, TRUE);
        gtk_widget_show_all(this_tim->candidate_window);
    } else {
        gtk_widget_hide(this_tim->candidate_window);
    }

    return handled;
}

static void gtkimcontexttim_focus_in(GtkIMContext *this)
{
    GtkIMContextTIM *this_tim = GTKIMCONTEXTTIM(this);

    if (GTKIMCONTEXTTIM(this)->debug) {
        printf("gtkimcontexttim_focus_in\n");
    }

    if (this_tim->showing) {
        gtk_widget_show_all(this_tim->candidate_window);
    }

    return;
}

static void gtkimcontexttim_focus_out(GtkIMContext *this)
{
    GtkIMContextTIM *this_tim = GTKIMCONTEXTTIM(this);

    if (GTKIMCONTEXTTIM(this)->debug) {
        printf("gtkimcontexttim_focus_out\n");
    }

    if (this_tim->showing) {
        gtk_widget_hide(this_tim->candidate_window);
    }

    return;
}

static void gtkimcontexttim_set_cursor_location(GtkIMContext *this,
                                                GdkRectangle *area)
{
    gint x, y;
    GtkIMContextTIM *this_tim = GTKIMCONTEXTTIM(this);

    if (GTKIMCONTEXTTIM(this)->debug) {
        printf("gtkimcontexttim_set_cursor_location: %d, %d, %d, %d\n",
               area->x, area->y, area->width, area->height);
    }

    if (this_tim->client_window != NULL) {
        gdk_window_get_origin(this_tim->client_window, &x, &y);
        candidate_window_move(this_tim,
                                   x + area->x, y + area->y,
                                   area->width, area->height);
    }

    return;
}

static void gtkimcontexttim_reset(GtkIMContext *this)
{
    if (GTKIMCONTEXTTIM(this)->debug) {
        printf("gtkimcontexttim_reset\n");
    }

    return;
}

static void gtkimcontexttim_get_preedit_string(GtkIMContext *this,
                                               gchar **str,
                                               PangoAttrList **attrs,
                                               gint *cursor_pos)
{
    register int i;
    GtkIMContextTIM *this_tim = GTKIMCONTEXTTIM(this);

    if (GTKIMCONTEXTTIM(this)->debug) {
        printf("gtkimcontexttim_get_preedit_string\n");
    }

    i = strutf8seq((guchar *)this_tim->candidates->str, &this_tim->which);
    if (i > 0) {
        guchar *p;

        i = -1;
        p = (guchar *)strchr((char *)this_tim->which, US);
        if (p == NULL) {
            p = (guchar *)strchr(this_tim->candidates->str, US);
            if (p == NULL) {
                i = strlen((char *)this_tim->which);
            }
        }
    }

    if (str) {
        GString *s;

        s = g_string_new(this_tim->converted->str);
        if (i > 0) {
            g_string_append_len(s, (char *)this_tim->which, i);
        } else {
            g_string_append(s, this_tim->raw->str);
        }
        *str = g_string_free(s, FALSE);
    }

    if (attrs) {
        PangoAttribute *attr;

        *attrs = pango_attr_list_new();

        /* underline */
        attr = pango_attr_underline_new(PANGO_UNDERLINE_SINGLE);
        attr->start_index = 0;
        attr->end_index = this_tim->converted->len;
        if (i > 0) {
            attr->end_index += i;
        } else {
            attr->end_index += this_tim->raw->len;;
        }
        pango_attr_list_insert(*attrs, attr);
        /* italic */
        attr = pango_attr_style_new(PANGO_STYLE_ITALIC);
        attr->start_index = this_tim->converted->len;
        attr->end_index = this_tim->converted->len;
        if (i > 0) {
            attr->end_index += i;
        } else {
            attr->end_index += this_tim->raw->len;;
        }
        pango_attr_list_insert(*attrs, attr);
    }

    if (cursor_pos) {
        *cursor_pos = this_tim->converted->len;
        if (i > 0) {
            *cursor_pos += i;
        } else {
            *cursor_pos += this_tim->raw->len;;
        }
    }
}

static void gtkimcontexttim_class_init(gpointer class, gpointer class_data)
{
#ifdef DEBUG
printf("gtkimcontexttim_class_init\n");
#endif
    /* GtkIMContextTIMClass specific methods initialization. */
    GTKIMCONTEXTTIM_CLASS(class)->set = gtkimcontexttim_set;

    /* Override super class methods. */
    GTK_IM_CONTEXT_CLASS(class)->set_client_window = gtkimcontexttim_set_client_window;
    GTK_IM_CONTEXT_CLASS(class)->filter_keypress = gtkimcontexttim_filter_keypress;
    GTK_IM_CONTEXT_CLASS(class)->reset = gtkimcontexttim_reset;
    GTK_IM_CONTEXT_CLASS(class)->get_preedit_string = gtkimcontexttim_get_preedit_string;
    GTK_IM_CONTEXT_CLASS(class)->focus_in = gtkimcontexttim_focus_in;
    GTK_IM_CONTEXT_CLASS(class)->focus_out = gtkimcontexttim_focus_out;
    GTK_IM_CONTEXT_CLASS(class)->set_cursor_location = gtkimcontexttim_set_cursor_location;

    /* Override GObjectClass finalize(). */
    G_OBJECT_CLASS(class)->finalize = gtkimcontexttim_finalize;
}

static GType gtkimcontexttim_type = 0;

GType gtkimcontexttim_get_type(void)
{
#ifdef DEBUG
printf("gtkimcontexttim_get_type\n");
#endif

    if (!gtkimcontexttim_type) {
        static const GTypeInfo gtkimcontexttim_info = {
            sizeof(GtkIMContextTIMClass),
            NULL,                       /* base_init */
            NULL,                       /* base_finalize */
            gtkimcontexttim_class_init,
            NULL,                       /* class_finalize */
            NULL,                       /* class_data */
            sizeof(GtkIMContextTIM),
            0,                          /* n_preallocs */
            gtkimcontexttim_init
        };

        gtkimcontexttim_type = g_type_register_static(
                GTK_TYPE_IM_CONTEXT,
                "GtkIMContextTIM",
                &gtkimcontexttim_info,
                0);
    }

    return gtkimcontexttim_type;
}

void gtkimcontexttim_register_type(GTypeModule *type_module)
{
#ifdef DEBUG
printf("gtkimcontexttim_register_type\n");
#endif

    if (!gtkimcontexttim_type) {
        static const GTypeInfo gtkimcontexttim_info = {
            sizeof(GtkIMContextTIMClass),
            NULL,                       /* base_init */
            NULL,                       /* base_finalize */
            gtkimcontexttim_class_init,
            NULL,                       /* class_finalize */
            NULL,                       /* class_data */
            sizeof(GtkIMContextTIM),
            0,                          /* n_preallocs */
            gtkimcontexttim_init
        };

        gtkimcontexttim_type = g_type_module_register_type(
                type_module,
                GTK_TYPE_IM_CONTEXT,
                "GtkIMContextTIM",
                &gtkimcontexttim_info,
                0);
    }
}
