/*
 * timtable.c - TIMTable class.
 *
    Copyright (C) 2002-2003  Yao Zhang

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * yaoz@users.sourceforge.net
 */
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "timtable.h"

#undef DEBUG

static int preprocess_line(char *line, int length)
{
    register int i;

    for (i = 0; i < length; i++) {
        switch (line[i]) {
        case '\\':
            /* escape */
            if (i < length-1) {
                length--;
                memmove(line+i, line+i+1, length-i);
            }
            break;
        case '#':
            /* comment */
            line[i] = '\0';
            length = i;
            break;
        case '=':
            /* record separator */
            line[i] = RS;    /* change it to <RS> */
            break;
        case '|':
            /* unit separator */
            line[i] = US;    /* change it to <US> */
            break;
        case '\0':
        case '\n':
            line[i] = '\0';
            length = i;
            break;
        default:
            break;
        }
    }

    return i;
}

void get_im_name(const char *dir, const char *entry, GString *name)
{
    FILE *fp;

    g_string_assign(name, dir);
    g_string_append(name, "/");
    g_string_append(name, entry);
#ifdef DEBUG
fprintf(stderr, "%s\n", name->str);
#endif

    fp = fopen(name->str, "r");
    g_string_assign(name, "");
    if (fp != NULL) {
        int l;
        size_t n = 512;
        char *line;

        line = (char *)malloc(n*sizeof(char));;
        /* getline() is a GNU externsion */
        while ((l = getline(&line, &n, fp)) > 0) {
            l = preprocess_line(line, l);
            if (l > 0) {
                char *p;

                p = memchr(line, RS, l);
                if ((p > line) && (*(p+1) != '\0')) {
                    char *k, *v;

                    *p = '\0';
                    k = line;
                    v = p+1;
		    if (strcmp(k, "TIMName") == 0) {
			g_string_assign(name, v);
#ifdef DEBUG
fprintf(stderr, "%s\n", name->str);
#endif
			break;
		    }
		}
	    }
	}
        free(line);
        fclose(fp);
    }
}

static void timtable_init(GTypeInstance *this, gpointer class)
{
    /* TIMTable specific fields initialization. */
#ifdef DEBUG
printf("timtable_init\n");
#endif
    TIMTABLE(this)->table = g_hash_table_new_full(
                                    (GHashFunc)g_string_hash,
                                    (GEqualFunc)g_string_equal,
                                    NULL,
                                    NULL);
}

static gboolean free_kv(gpointer key, gpointer value, gpointer user_data)
{
    g_string_free(key, TRUE);
    g_string_free(value, TRUE);

    return TRUE;
}

static void timtable_finalize(GObject *this)
{
    GObjectClass *super_class;

    /* TIMTable specific finalization. */
#ifdef DEBUG
printf("timtable_finalize\n");
#endif
    g_hash_table_foreach_remove(TIMTABLE(this)->table,
		                free_kv, NULL);
    g_hash_table_destroy(TIMTABLE(this)->table);

    /* Super class finalizatoin. */
    super_class = g_type_class_peek(g_type_parent(TIMTABLE_TYPE));
    G_OBJECT_CLASS(super_class)->finalize(this);
}

static void timtable_load(TIMTable *this, gchar *file_name)
{
    GString *file;
    FILE *fp;

    file = g_string_new("table.tim");
    if ((file_name != NULL) && (strlen(file_name) > 0)) {
        g_string_assign(file, file_name);
    }

    fp = fopen(file->str, "r");
    if (fp == NULL) {
	if (file->str[0] != '/') {
	    GString *path;

	    path = g_string_new(getenv("HOME"));
            g_string_append(path, "/.wenju/");
            g_string_append(path, file->str);
            fp = fopen(path->str, "r");
	    if (fp == NULL) {
		g_string_assign(path, PKGDATADIR);
                g_string_append(path, "/");
                g_string_append(path, file->str);
                fp = fopen(path->str, "r");
	    }
            g_string_free(path, TRUE);
	}
    }

    g_string_free(file, TRUE);

    if (fp != NULL) {
        int l;
        size_t n = 512;
        char *line;

        line = (char *)malloc(n*sizeof(char));;
        /* getline() is a GNU externsion */
        while ((l = getline(&line, &n, fp)) > 0) {
            l = preprocess_line(line, l);
            if (l > 0) {
                char *p;

                p = memchr(line, RS, l);
                if ((p > line) && (*(p+1) != '\0')) {
                    char *k, *v;

                    *p = '\0';
                    k = line;
                    v = p+1;
                    if (strcmp(k, "Include") == 0) {
                        if (v[0] != '\0') {
                            timtable_load(this, v);
                        }
                    } else {
                        GString *key, *value;

                        key = g_string_new(k);
                        value = (GString *)g_hash_table_lookup(this->table,
                                                               key);
                        if (value == NULL) {
                            value = g_string_new(v);
                            g_hash_table_insert(this->table, key, value);
                        } else {
                            g_string_free(key, TRUE);
                            g_string_append_c(value, US);
                            g_string_append(value, v);
                        }
#ifdef DEBUG
key = g_string_new(k);
value = (GString *)g_hash_table_lookup(this->table, key);
printf("key: %s, value: %s\n", key->str, value->str);
g_string_free(key, TRUE);
#endif
                    }
                }
            }
        }
        free(line);
        fclose(fp);
    }
}

static GString *timtable_lookup(TIMTable *this, GString *key)
{
    return (GString *)g_hash_table_lookup(this->table, key);
}

static void timtable_class_init(gpointer class, gpointer class_data)
{
    /* TIMTableClass specific methods initialization. */
    TIMTABLE_CLASS(class)->load = timtable_load;
    TIMTABLE_CLASS(class)->lookup = timtable_lookup;

    /* Override GObjectClass finalize(). */
    G_OBJECT_CLASS(class)->finalize = timtable_finalize;
}

static GType timtable_type = 0;

GType timtable_get_type(void)
{
    if (!timtable_type) {
        static const GTypeInfo timtable_info = {
            sizeof(TIMTableClass),          /* class_size */
            NULL,                       /* base_init */
            NULL,                       /* base_finalize */
            timtable_class_init,            /* class_init */
            NULL,                       /* class_finalize */
            NULL,                       /* class_data */
            sizeof(TIMTable),               /* instance_size */
            0,                          /* n_preallocs */
            timtable_init                   /* instance_init */
        };

        timtable_type = g_type_register_static(G_TYPE_OBJECT,
                                           "TIMTable",
                                           &timtable_info,
                                           0);
    }

    return timtable_type;
}

void timtable_register_type(GTypeModule *type_module)
{
    if (!timtable_type) {
        static const GTypeInfo timtable_info = {
            sizeof(TIMTableClass),          /* class_size */
            NULL,                       /* base_init */
            NULL,                       /* base_finalize */
            timtable_class_init,            /* class_init */
            NULL,                       /* class_finalize */
            NULL,                       /* class_data */
            sizeof(TIMTable),               /* instance_size */
            0,                          /* n_preallocs */
            timtable_init                   /* instance_init */
        };

        timtable_type = g_type_module_register_type(type_module,
                                           G_TYPE_OBJECT,
                                           "TIMTable",
                                           &timtable_info,
                                           0);
    }
}
