/*
    Copyright (C) 2002-2003  Yao Zhang

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * yaoz@users.sourceforge.net
 */
/**
 * wrap.c
 *
 * It is a small utility to wrap around the input text file so that
 * the line width is about the same.  The default line width is 
 * 64 columns.  It knows about the width of a character (for example,
 * ASCII characters are 1 column where Chinese characters are 2 
 * columns) and tries to break the line only at appropriate boundaries 
 * (like after a space ' ' or after a punctuation, etc.).  An empty 
 * line by itself signals the end of the paragraph.
 */
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>

static int utf8_width(gchar *utf8)
{
    extern int mk_wcwidth(wchar_t ucs);

    int n = mk_wcwidth(g_utf8_get_char(utf8));
    return n >= 0 ? n : 0;
}

static int utf8_isalnum(gchar *utf8)
{
    return (((*utf8 >= '0') && (*utf8 <= '9'))
         || ((*utf8 >= 'A') && (*utf8 <= 'Z'))
         || ((*utf8 >= 'a') && (*utf8 <= 'z')));
}

static int is_breakable(GString *text, int cursor)
{
    gchar *p;

    p = g_utf8_find_prev_char(text->str, text->str + cursor);
    /* Characters before and after cursor are alnums, non breakable. */
    if ((p != NULL) && utf8_isalnum(p)
     && utf8_isalnum(text->str + cursor)) {
        return 0;
    }

    /* Character before cursor is a start quote, etc., non breakable. */
    if (p != NULL) {
        if (strncmp(p, "(", strlen("(")) == 0) {
            return 0;
        } else if (strncmp(p, "<", strlen("<")) == 0) {
            return 0;
        } else if (strncmp(p, "[", strlen("[")) == 0) {
            return 0;
        } else if (strncmp(p, "‘", strlen("‘")) == 0) {
            return 0;
        } else if (strncmp(p, "“", strlen("“")) == 0) {
            return 0;
        } else if (strncmp(p, "（", strlen("（")) == 0) {
            return 0;
        } else if (strncmp(p, "〈", strlen("〈")) == 0) {
            return 0;
        } else if (strncmp(p, "《", strlen("《")) == 0) {
            return 0;
        } else if (strncmp(p, "_", strlen("_")) == 0) {
            return 0;
        }
    }

    /* Character after cursor is a space, etc., non breakable. */
    p = text->str + cursor;
    if (*p != '\0') {
	if (strncmp(p, " ", strlen(" ")) == 0) {
            return 0;
        } else if (strncmp(p, ",", strlen(",")) == 0) {
            return 0;
        } else if (strncmp(p, ".", strlen(".")) == 0) {
            return 0;
        } else if (strncmp(p, ",", strlen(",")) == 0) {
            return 0;
        } else if (strncmp(p, "?", strlen("?")) == 0) {
            return 0;
        } else if (strncmp(p, "!", strlen("!")) == 0) {
            return 0;
        } else if (strncmp(p, ";", strlen(";")) == 0) {
            return 0;
        } else if (strncmp(p, ":", strlen(":")) == 0) {
            return 0;
        } else if (strncmp(p, ")", strlen(")")) == 0) {
            return 0;
        } else if (strncmp(p, ">", strlen(">")) == 0) {
            return 0;
        } else if (strncmp(p, "]", strlen("]")) == 0) {
            return 0;
        } else if (strncmp(p, "’", strlen("’")) == 0) {
            return 0;
        } else if (strncmp(p, "”", strlen("”")) == 0) {
            return 0;
        } else if (strncmp(p, "、", strlen("、")) == 0) {
            return 0;
        } else if (strncmp(p, "，", strlen("，")) == 0) {
            return 0;
        } else if (strncmp(p, "。", strlen("。")) == 0) {
            return 0;
        } else if (strncmp(p, "？", strlen("？")) == 0) {
            return 0;
        } else if (strncmp(p, "；", strlen("；")) == 0) {
            return 0;
        } else if (strncmp(p, "）", strlen("）")) == 0) {
            return 0;
        } else if (strncmp(p, "〉", strlen("〉")) == 0) {
            return 0;
        } else if (strncmp(p, "》", strlen("》")) == 0) {
            return 0;
        } else if (strncmp(p, "_", strlen("_")) == 0) {
            return 0;
        }
    }

    return 1;
}

/**
 * Find the closest breaking point to the columns width.
 * If the whole text is less than columns, it returns -1; otherwise,
 * returns the cursor position to the breaking point.
 */
static int breaking_point(GString *text, int columns)
{
    int w, wp, wq;
    gchar *n, *p, *q;

    w = 0;
    n = text->str;
    while (n != NULL) {
        w += utf8_width(n);
        if (w >= columns) {
            break;
        }
	n = g_utf8_find_next_char(n, text->str + text->len);
    }

    if (n == NULL) {
	/* w < columns */
        return -1;
    } else {
        /**
	 * Move the pointer after the last character so that the cursor
	 * position is the offset between the pointer and the start of the
	 * string.  Function is_breakable() uses this cursor position.
	 */
        n = g_utf8_find_next_char(n, text->str + text->len);

	/* Try to find the breaking point before the cursor. */
        wp = w;
        p = g_utf8_find_prev_char(text->str, n);
        while (p != NULL) {
            wp -= utf8_width(p);
            if (is_breakable(text, p - text->str)) {
                break;
            }
            p = g_utf8_find_prev_char(text->str, p);
        }

	/* Try to find the breaking point on or after the cursor. */
        wq = w;
        q = n;
        while (q != NULL) {
            if (is_breakable(text, q - text->str)) {
                break;
            }
            wq += utf8_width(q);
            q = g_utf8_find_next_char(q, text->str + text->len);
        }

	/* Favors the shorter line. */
        if ((columns-wp) <= (wq-columns)) {
            return p - text->str;
        }

        return q - text->str;
    }
}

/**
 * If there are '\n' in the middle of text, change it to ' ' if
 * it is surrounded by English text, otherwise remove it.
 */
static void handle_middle_newline(GString *text)
{
    gchar *n;

    n = text->str;
    while ((n = strchr(n, '\n')) != NULL) {
        gchar *p, *q;

	/* Do not change the last '\n'. */
        if (n[1] == '\0') {
            break;
        }

        p = g_utf8_find_prev_char(text->str, n);
        q = g_utf8_find_next_char(n, text->str + text->len);
        if ((p != NULL) && utf8_isalnum(p)
         && (q != NULL) && utf8_isalnum(q)) {
            *n = ' ';
        } else {
            g_string_erase(text, n - text->str, 1);
        }
    }
}

int main(int argc, char *argv[])
{
    register int i;
    int columns;
    GString *text;
    FILE *fp;

    columns = 64;

    while (1) {
        int c;
        static struct option longopts[] = {
            {"width", required_argument, NULL, 'w'},
            {"help", no_argument, NULL, 'h'},
            {NULL, 0, NULL, 0}
        };
        int longindex;

        c = getopt_long(argc, argv, "w:",
                        longopts, &longindex);
        if (c == -1) {
            break;
        }

        switch (c) {
        case 'w':
            columns = atoi(optarg);
            break;
        case 'h':
            printf("Usage: wrap [OPTION...]\n");
            printf("Wrap each input line to fit in specified width.\n");
            printf("\n");
            printf("  -w, --width=WIDTH    use WIDTH columns instead of 64\n");
            printf("      --help           display this help and exit\n");
            exit(0);
            break;
        case '?':
	    /* Unknown option. */
            break;
        default:
	    /* Shouldn't happen. */
            printf("?? getopt returned character code 0x%X ??\n", c);
            break;
        }
    }

    if (argc-optind > 0) {
        /* Too many arguments. */
        printf("Usage: wrap [OPTION...]\n");
        printf("Wrap each input line to fit in specified width.\n");
        printf("\n");
        printf("  -w, --width=WIDTH    use WIDTH columns instead of 64\n");
        printf("      --help           display this help and exit\n");
        exit(1);
    }

    if (columns < 3) {
        columns = 64;
    }

    /**
     * The UString 'text' holds the text to be broken into lines
     * about the same columns long.
     */
    text = g_string_new("");

    fp = fopen("/dev/stdin", "r");
    if (fp != NULL) {
        int length;
        size_t size;
        char *p;

        size = 0;
        p = NULL;
        /* Read in an input line. */
        while ((length = getline(&p, &size, fp)) > 0) {
            if (p[0] == '\n') {
                /* An empty line signals the end of the paragraph. */
                if (strcmp(text->str, "\n") == 0) {
                    g_string_assign(text, "");
                }
                printf("%s\n", text->str);
                g_string_assign(text, "");
            } else {
                /* Append the input line to the text. */
                g_string_append(text, p);

                /**
                 * There could be '\n' in the middle of the text.
		 * It should be either removed or changed to ' '.
                 */
                handle_middle_newline(text);

                /* Find the right breaking point. */
                while ((i = breaking_point(text, columns)) >= 0) {
                    printf("%.*s\n", i, text->str);
                    g_string_erase(text, 0, i);
                }
	    }
        }
        if (strcmp(text->str, "\n") == 0) {
            g_string_assign(text, "");
        }
        printf("%s", text->str);
        g_string_assign(text, "");
        if (p != NULL) {
            free(p);
        }

        fclose(fp);
    }

    g_string_free(text, TRUE);

    exit(0);
}
